from typing import Any, Dict, Hashable, Mapping, TypeVar, overload

from pendulum import Date, now

__all__ = ("mapping_merge", "today")

Q = TypeVar("Q", bound=Hashable)
T = TypeVar("T")


@overload
def mapping_merge(*mappings: Mapping[str, T], **keywords: T) -> Dict[str, T]:
    ...


@overload
def mapping_merge(*mappings: Mapping[Q, T]) -> Dict[Q, T]:
    ...


def mapping_merge(*mappings: Mapping[Any, Any], **keywords: Any) -> Dict[Any, Any]:
    """Merges `mappings` and `keywords` into one dictionary.

    Arguments:
        *mappings: Mappings to merge.
        **keywords: Keywords to add to the result.

    Returns:
        A newly created dictionary.
    """
    result: Dict[Any, Any] = {}

    for mapping in mappings:
        result.update(mapping)

    result.update(keywords)

    return result


def today() -> Date:
    """Returns the current date."""
    return now().date()  # type: ignore



from os import PathLike
from typing import Any, Callable, Dict, Tuple, TypeVar, Union

from typing_extensions import TypeAlias

__all__ = ("AnyException", "DynamicTuple", "StringDict", "AnyStringDict", "Nullary", "Unary")

T = TypeVar("T")
R = TypeVar("R")

AnyException: TypeAlias = BaseException

DynamicTuple = Tuple[T, ...]

StringDict = Dict[str, T]

AnyStringDict = StringDict[Any]

Nullary = Callable[[], R]
Unary = Callable[[T], R]

try:  # pragma: no cover
    IntoPath = Union[PathLike[str], str]  # type: ignore

except TypeError:  # pragma: no cover
    IntoPath = Union[PathLike, str]  # type: ignore



from pathlib import Path
from subprocess import call
from typing import Iterable, Sequence

from changelogging.typing import DynamicTuple

__all__ = ("remove_command", "remove_paths")

GIT = "git"
REMOVE = "rm"
QUIET = "--quiet"
FORCE = "--force"


def remove_command(iterable: Iterable[Path]) -> Sequence[str]:
    command: DynamicTuple[str] = (GIT, REMOVE, QUIET, FORCE)

    command += tuple(path.resolve().as_posix() for path in iterable)

    return command


def remove_paths(iterable: Iterable[Path]) -> None:
    call(remove_command(iterable))



from collections import defaultdict as default_dict
from pathlib import Path
from textwrap import wrap
from typing import Dict, Iterable, Iterator, List, Mapping, Tuple, TypeVar

from attrs import define, field
from pendulum import Date, now
from typing_extensions import Literal

from changelogging.config import Config
from changelogging.constants import DOT, DOUBLE_NEW_LINE, EMPTY, HASH, NEW_LINE, SPACE
from changelogging.fragments import Fragment, FragmentType, Issue

__all__ = ("Builder",)

NO_SIGNIFICANT_CHANGES = "No significant changes."

concat_new_line = NEW_LINE.join
concat_double_new_line = DOUBLE_NEW_LINE.join

WRITE: Literal["w"] = "w"


def today() -> Date:
    return now().date()  # type: ignore


FT = TypeVar("FT", bound=FragmentType)
IT = TypeVar("IT", bound=Issue)

Fragments = Iterable[Fragment[FT, IT]]
Sections = Mapping[FT, Fragments[FT, IT]]


@define()
class Builder:
    """Represents changelog builders."""

    config: Config = field()
    """The config of the builder."""

    date: Date = field(factory=today)
    """The date to use in builds."""

    def build_title(self) -> str:
        """Builds the main title.

        Returns:
            The main title.
        """
        config = self.config

        return self.heading(config.title_level) + config.title_format.format(
            name=config.name, version=config.version, url=config.url, date=self.date
        )

    def build_section_title(self, type: FragmentType) -> str:
        """Builds the section title of the particular `type`.

        Arguments:
            type: The fragment type to build the title for.

        Returns:
            The section title.
        """
        config = self.config

        return self.heading(config.section_level) + type.title

    def build_issue(self, issue: Issue) -> str:
        """Builds the `issue` given.

        Arguments:
            issue: The issue to build.

        Returns:
            The issue built.
        """
        config = self.config

        return config.issue_format.format(issue=issue.value, url=config.url)

    def build_fragment(self, fragment: Fragment[FragmentType, Issue]) -> str:
        """Builds the `fragment` given.

        Arguments:
            fragment: The fragment to build.

        Returns:
            The built fragment.
        """
        config = self.config

        content = config.fragment_format.format(
            content=fragment.content.strip(), issue=self.build_issue(fragment.issue)
        )

        return (
            self.indent_wrap_lines(content, config.bullet, config.wrap_size)
            if config.wrap
            else self.indent_lines(content, config.bullet)
        )

    def build_fragments(self, fragments: Fragments[FragmentType, Issue]) -> str:
        """Builds several `fragments` given.

        Arguments:
            fragments: The fragments to build.

        Returns:
            The fragments built.
        """
        return concat_new_line(map(self.build_fragment, fragments))

    def collect_sections(self, fragments: Fragments[FT, IT]) -> Sections[FT, IT]:
        """Collects `fragments` into sections.

        Arguments:
            fragments: The fragments to collect.

        Returns:
            The collected sections.
        """
        sections: Dict[FT, List[Fragment[FT, IT]]] = default_dict(list)

        for fragment in fragments:
            sections[fragment.type].append(fragment)

        for fragments in sections.values():
            fragments.sort()

        return sections

    def build_generate(self, sections: Sections[FragmentType, Issue]) -> Iterator[str]:
        """Builds `sections`, returning an iterator.

        Arguments:
            sections: The sections to build.

        Returns:
            The iterator over the build result.
        """
        config = self.config

        yield self.build_title()

        empty = True

        for type in config.display.into_types(config.types):
            if type in sections:
                empty = False

                fragments = sections[type]

                yield self.build_section_title(type)
                yield self.build_fragments(fragments)

        if empty:  # pragma: no cover  # not tested
            yield NO_SIGNIFICANT_CHANGES

    def collect_fragments(self) -> Fragments[FragmentType, Issue]:
        """Collects fragments from the changes directory specified in the config.

        Returns:
            The iterator over the fragments found.
        """
        for path, type in self.collect_paths_types():
            issue = self.get_issue(path)
            content = path.read_text()

            yield Fragment(type, content, issue)

    def collect_paths_types(self) -> Iterator[Tuple[Path, FragmentType]]:
        config = self.config
        directory = config.directory
        types = config.types

        for path in directory.iterdir():
            if path.is_file():
                for suffix in path.suffixes:
                    if types.has_suffix(suffix):
                        yield (path, types.get_suffix(suffix))

    def collect_paths(self) -> Iterator[Path]:  # pragma: no cover  # only used in `main`
        """Collect paths to fragments.

        Returns:
            The iterator over paths found.
        """
        for path, _ in self.collect_paths_types():
            yield path

    @classmethod
    def get_issue(cls, path: Path) -> Issue:
        return Issue(int(cls.name_no_suffixes(path)))

    def build(self) -> str:
        """Builds the changelog.

        Returns:
            The build result.
        """
        return concat_double_new_line(
            self.build_generate(self.collect_sections(self.collect_fragments()))
        )

    def write(self) -> None:
        """Builds the changelog and writes it to the output file."""
        config = self.config

        output = config.output
        start_string = config.start_string

        content = self.build()

        current = output.read_text() if output.exists() else EMPTY

        before, start, after = current.partition(start_string)

        with output.open(WRITE) as file:
            if not start:  # pragma: no cover  # not tested
                file.write(content + NEW_LINE)

                if current.strip():
                    file.write(NEW_LINE + current.lstrip())

            else:
                file.write(before)
                file.write(start)
                file.write(DOUBLE_NEW_LINE + content + NEW_LINE)

                if after.strip():
                    file.write(NEW_LINE + after.lstrip())

    @staticmethod
    def heading(level: int) -> str:
        return HASH * level + SPACE

    @staticmethod
    def indents(bullet: str) -> Tuple[str, str]:
        space = SPACE
        return (bullet + space, space + space)

    @staticmethod
    def name_no_suffixes(path: Path) -> str:
        name, _, _ = path.name.partition(DOT)
        return name

    @classmethod
    def generate_indent_lines(cls, string: str, bullet: str) -> Iterator[str]:
        if not string:  # pragma: no cover  # not tested
            return

        initial, subsequent = cls.indents(bullet)

        head, *tail = string.splitlines()

        yield (initial + head if head.strip() else head)

        for item in tail:
            yield (subsequent + item if item.strip() else item)

    @classmethod
    def indent_lines(cls, string: str, bullet: str) -> str:
        return concat_new_line(cls.generate_indent_lines(string, bullet))

    @classmethod
    def indent_wrap_lines(cls, string: str, bullet: str, size: int) -> str:
        size -= max(map(len, cls.indents(bullet)))

        return cls.indent_lines(
            concat_new_line(
                concat_new_line(wrap(line, size, break_long_words=False))
                for line in string.splitlines()
            ),
            bullet,
        )



from pathlib import Path
from typing import Any, Dict, Iterable, Optional, Type, TypeVar, cast, overload

import toml
from attrs import define
from iters import iter
from iters.utils import empty
from versions import Version, parse_version
from wraps import Option, wrap_optional
from yarl import URL

from changelogging.constants import (
    DEFAULT_ENCODING,
    DEFAULT_ERRORS,
    DEFAULT_IGNORE_REQUIRED,
    EMPTY,
    ROOT,
)
from changelogging.fragments import AnyFragmentTypes, Display, FragmentType
from changelogging.typing import IntoPath, Unary

__all__ = ("Config", "ConfigData", "AnyConfigData")

CHANGELOGGING = "changelogging.toml"
PYPROJECT = "pyproject.toml"

SEARCH = (CHANGELOGGING, PYPROJECT)

DEFAULT_PATH = ROOT / CHANGELOGGING

CONFIG_NOT_FOUND = "can not find config in {}"
EXPECTED_FILE_OR_DIRECTORY = "expected either a file or a directory"

EXPECTED = "expected `{}`"
expected = EXPECTED.format

EXPECTED_CHANGELOGGING = expected("changelogging")
EXPECTED_CHANGELOGGING_NAME = expected("changelogging.name")
EXPECTED_CHANGELOGGING_VERSION = expected("changelogging.version")
EXPECTED_CHANGELOGGING_URL = expected("changelogging.url")
EXPECTED_CHANGELOGGING_DIRECTORY = expected("changelogging.directory")
EXPECTED_CHANGELOGGING_OUTPUT = expected("changelogging.output")
EXPECTED_CHANGELOGGING_TITLE_LEVEL = expected("changelogging.title_level")
EXPECTED_CHANGELOGGING_SECTION_LEVEL = expected("changelogging.section_level")
EXPECTED_CHANGELOGGING_BULLET = expected("changelogging.bullet")
EXPECTED_CHANGELOGGING_WRAP = expected("changelogging.wrap")
EXPECTED_CHANGELOGGING_WRAP_SIZE = expected("changelogging.wrap_size")
EXPECTED_CHANGELOGGING_START_STRING = expected("changelogging.start_string")
EXPECTED_CHANGELOGGING_TITLE_FORMAT = expected("changelogging.title_format")
EXPECTED_CHANGELOGGING_ISSUE_FORMAT = expected("changelogging.issue_format")
EXPECTED_CHANGELOGGING_FRAGMENT_FORMAT = expected("changelogging.fragment_format")
EXPECTED_CHANGELOGGING_DISPLAY = expected("changelogging.display")
EXPECTED_CHANGELOGGING_TYPES_TYPE_NAME = expected("changelogging.types.type.name")
EXPECTED_CHANGELOGGING_TYPES_TYPE_TITLE = expected("changelogging.types.type.title")


def config_not_found(path: Path) -> FileNotFoundError:
    return FileNotFoundError(CONFIG_NOT_FOUND.format(path.resolve().as_posix()))


def expected_file_or_directory() -> ValueError:  # pragma: no cover
    return ValueError(EXPECTED_FILE_OR_DIRECTORY)


T = TypeVar("T")


CD = TypeVar("CD", bound="AnyConfigData")


class ConfigData(Dict[str, T]):
    """Dictionaries that support attribute access."""

    def __getattr__(self, name: str) -> Option[T]:
        return wrap_optional(self.get(name))

    def copy(self: CD) -> CD:
        return type(self)(self)


AnyConfigData = ConfigData[Any]


FT = TypeVar("FT", bound=FragmentType)


@overload
def mapping_to_type(mapping: ConfigData[str]) -> FragmentType:
    ...


@overload
def mapping_to_type(mapping: ConfigData[str], fragment_type: Type[FT]) -> FT:
    ...


def mapping_to_type(mapping: ConfigData[str], fragment_type: Type[Any] = FragmentType) -> Any:
    return fragment_type(
        mapping.name.expect(EXPECTED_CHANGELOGGING_TYPES_TYPE_NAME),
        mapping.title.expect(EXPECTED_CHANGELOGGING_TYPES_TYPE_TITLE),
    )


C = TypeVar("C", bound="Config")


@define()
class Config:
    name: str
    """The name of the project."""
    version: Version
    """The version of the project."""
    url: URL
    """The URL of the project."""
    directory: Path
    """The `changes` directory."""
    output: Path
    """The output path."""
    title_level: int
    """The title level to use."""
    section_level: int
    """The section title level to use."""
    bullet: str
    """The bullet to use."""
    wrap: bool
    """Whether to wrap lines."""
    wrap_size: int
    """The wrap size to use."""
    start_string: str
    """The start string to look for."""
    title_format: str
    """The format of the title."""
    issue_format: str
    """The format of the issue."""
    fragment_format: str
    """The format of the fragment."""
    display: Display
    """The display ordering of fragment types."""
    types: AnyFragmentTypes
    """The fragment types to use."""

    # dynamic code ahead...

    @classmethod
    def from_string(cls: Type[C], string: str, source: Optional[Path] = None) -> C:
        """Parses a [`Config`][changelogging.config.Config] from `string`.

        Arguments:
            string: The string to parse.
            source: The source of where the config came from.

        Returns:
            The newly parsed [`Config`][changelogging.config.Config].
        """
        return cls.from_data(cls.parse(string), source)

    @classmethod
    def from_file_path(
        cls: Type[C], path: IntoPath, encoding: str = DEFAULT_ENCODING, errors: str = DEFAULT_ERRORS
    ) -> C:
        """Parses a [`Config`][changelogging.config.Config] from file `path`.

        Arguments:
            path: The path to the config.
            encoding: The encoding to use.
            errors: The error handling strategy to use.

        Returns:
            The newly parsed [`Config`][changelogging.config.Config] instance.
        """
        path = Path(path)

        return cls.from_string(path.read_text(encoding, errors), path)

    @classmethod
    def from_path(
        cls: Type[C],
        path: IntoPath,
        search: Iterable[IntoPath] = SEARCH,
        encoding: str = DEFAULT_ENCODING,
        errors: str = DEFAULT_ERRORS,
    ) -> C:
        """Parses a [`Config`][changelogging.config.Config] from `path`.

        If `path` is a directory, this function searches for files in `search` inside of it.

        Arguments:
            path: The path to the config.
            search: The paths to search for.
            encoding: The encoding to use.
            errors: The error handling strategy to use.

        Returns:
            The newly parsed [`Config`][changelogging.config.Config].
        """
        path = Path(path)

        if not path.exists():
            raise config_not_found(path)

        if path.is_dir():
            for part in search:
                try_path = path / part

                if try_path.exists():
                    return cls.from_file_path(try_path, encoding, errors)

            raise config_not_found(path)

        if path.is_file():
            return cls.from_file_path(path, encoding, errors)

        raise expected_file_or_directory()  # pragma: no cover

    @staticmethod
    def parse(string: str) -> AnyConfigData:
        return cast(AnyConfigData, toml.loads(string, AnyConfigData))

    @classmethod
    def from_data(cls: Type[C], config_data: AnyConfigData, source: Optional[Path] = None) -> C:
        """Creates a [`Config`][changelogging.config.Config]
        from [`ConfigData`][changelogging.config.ConfigData].

        Arguments:
            config_data: The config data to use.

        Returns:
            The newly created [`Config`][changelogging.config.Config] instance.
        """
        config_data = config_data.tool.unwrap_or(config_data)
        config = config_data.changelogging.expect(EXPECTED_CHANGELOGGING)

        types = (
            iter(config.types.unwrap_or_else(empty))
            .map(mapping_to_type)
            .collect(AnyFragmentTypes.from_iterable)
        )

        default_config = DEFAULT_CONFIG

        return cls(
            # `name`, `version` and `url` are always required
            name=config.name.expect(EXPECTED_CHANGELOGGING_NAME),
            version=config.version.map(parse_version).expect(EXPECTED_CHANGELOGGING_VERSION),
            # map to `URL` and `Path` to simplify interaction
            url=config.url.map(URL).expect(EXPECTED_CHANGELOGGING_URL),
            directory=config.directory.map_or(default_config.directory, source_path(source)),
            output=config.output.map_or(default_config.output, source_path(source)),
            # merely return defaults if needed
            title_level=config.title_level.unwrap_or(default_config.title_level),
            section_level=config.section_level.unwrap_or(default_config.section_level),
            bullet=config.bullet.unwrap_or(default_config.bullet),
            wrap=config.wrap.unwrap_or(default_config.wrap),
            wrap_size=config.wrap_size.unwrap_or(default_config.wrap_size),
            start_string=config.start_string.unwrap_or(default_config.start_string),
            title_format=config.title_format.unwrap_or(default_config.title_format),
            issue_format=config.issue_format.unwrap_or(default_config.issue_format),
            fragment_format=config.fragment_format.unwrap_or(default_config.fragment_format),
            # map to `Display` type
            display=config.display.map_or(default_config.display, Display.from_iterable),
            # merge user-defined `types` with already existing default ones
            types=default_config.types.merge_with(types),
        )

    @classmethod
    def unsafe_from_string(
        cls: Type[C],
        string: str,
        source: Optional[Path] = None,
        ignore_required: bool = DEFAULT_IGNORE_REQUIRED,
    ) -> C:
        return cls.unsafe_from_data(cls.parse(string), source, ignore_required=ignore_required)

    @classmethod
    def unsafe_from_file_path(
        cls: Type[C],
        path: IntoPath,
        encoding: str = DEFAULT_ENCODING,
        errors: str = DEFAULT_ERRORS,
        ignore_required: bool = DEFAULT_IGNORE_REQUIRED,
    ) -> C:
        path = Path(path)

        return cls.unsafe_from_string(
            path.read_text(encoding, errors), path, ignore_required=ignore_required
        )

    @classmethod
    def unsafe_from_path(
        cls: Type[C],
        path: IntoPath,
        search: Iterable[IntoPath] = SEARCH,
        encoding: str = DEFAULT_ENCODING,
        errors: str = DEFAULT_ERRORS,
        ignore_required: bool = DEFAULT_IGNORE_REQUIRED,
    ) -> C:
        path = Path(path)

        if not path.exists():  # pragma: no cover
            raise config_not_found(path)

        if path.is_dir():  # pragma: no cover
            for part in search:
                try_path = path / part

                if try_path.exists():
                    return cls.unsafe_from_file_path(
                        try_path, encoding, errors, ignore_required=ignore_required
                    )

            raise config_not_found(path)

        if path.is_file():
            return cls.unsafe_from_file_path(
                path, encoding, errors, ignore_required=ignore_required
            )

        raise expected_file_or_directory()  # pragma: no cover

    @classmethod
    def unsafe_from_data(
        cls: Type[C],
        config_dict: AnyConfigData,
        source: Optional[Path] = None,
        ignore_required: bool = DEFAULT_IGNORE_REQUIRED,
    ) -> C:
        config_dict = config_dict.tool.unwrap_or(config_dict)
        config = config_dict.changelogging.expect(EXPECTED_CHANGELOGGING)

        types = (
            iter(config.types.unwrap_or_else(empty))
            .map(mapping_to_type)
            .collect(AnyFragmentTypes.from_iterable)
        )

        name_option = config.name

        name = (
            name_option.unwrap_or(EMPTY)
            if ignore_required
            else name_option.expect(EXPECTED_CHANGELOGGING_NAME)
        )

        version_option = config.version.map(parse_version)

        version = (
            version_option.unwrap_or_else(Version)
            if ignore_required
            else version_option.expect(EXPECTED_CHANGELOGGING_VERSION)
        )

        url_option = config.url.map(URL)

        url = (
            url_option.unwrap_or_else(URL)
            if ignore_required
            else url_option.expect(EXPECTED_CHANGELOGGING_URL)
        )

        return cls(
            name=name,
            version=version,
            url=url,
            directory=config.directory.map(source_path(source)).expect(
                EXPECTED_CHANGELOGGING_DIRECTORY
            ),
            output=config.output.map(source_path(source)).expect(EXPECTED_CHANGELOGGING_OUTPUT),
            title_level=config.title_level.expect(EXPECTED_CHANGELOGGING_TITLE_LEVEL),
            section_level=config.section_level.expect(EXPECTED_CHANGELOGGING_SECTION_LEVEL),
            bullet=config.bullet.expect(EXPECTED_CHANGELOGGING_BULLET),
            wrap=config.wrap.expect(EXPECTED_CHANGELOGGING_WRAP),
            wrap_size=config.wrap_size.expect(EXPECTED_CHANGELOGGING_WRAP_SIZE),
            start_string=config.start_string.expect(EXPECTED_CHANGELOGGING_START_STRING),
            title_format=config.title_format.expect(EXPECTED_CHANGELOGGING_TITLE_FORMAT),
            issue_format=config.issue_format.expect(EXPECTED_CHANGELOGGING_ISSUE_FORMAT),
            fragment_format=config.fragment_format.expect(EXPECTED_CHANGELOGGING_FRAGMENT_FORMAT),
            display=config.display.map(Display.from_iterable).expect(
                EXPECTED_CHANGELOGGING_DISPLAY
            ),
            types=types,
        )


def source_path(source: Optional[Path] = None) -> Unary[str, Path]:
    def format_path(string: str) -> Path:
        if source is None:
            return Path(string)

        return Path(string.format(here=source.parent))

    return format_path


DEFAULT_CONFIG = Config.unsafe_from_path(DEFAULT_PATH, ignore_required=True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['changelogging']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=22.2.0',
 'click>=8.1.3',
 'entrypoint>=1.3.0',
 'pendulum>=2.1.2',
 'toml>=0.10.2',
 'typing-extensions>=4.3.0',
 'versions>=1.3.0',
 'wraps>=0.3.0',
 'yarl>=1.8.2']

entry_points = \
{'console_scripts': ['changelogging = changelogging.main:changelogging']}

setup_kwargs = {
    'name': 'changelogging',
    'version': '1.1.0',
    'description': 'Building changelogs from fragments.',
    'long_description': '# `changelogging`\n\n[![License][License Badge]][License]\n[![Version][Version Badge]][Package]\n[![Downloads][Downloads Badge]][Package]\n[![Discord][Discord Badge]][Discord]\n\n[![Documentation][Documentation Badge]][Documentation]\n[![Check][Check Badge]][Actions]\n[![Test][Test Badge]][Actions]\n[![Coverage][Coverage Badge]][Coverage]\n\n> *Building changelogs from fragments.*\n\n## Installing\n\n**Python 3.7 or above is required.**\n\n### pip\n\nInstalling the library with `pip` is quite simple:\n\n```console\n$ pip install changelogging\n```\n\nAlternatively, the library can be installed from source:\n\n```console\n$ git clone https://github.com/nekitdev/changelogging.git\n$ cd changelogging\n$ python -m pip install .\n```\n\n### poetry\n\nYou can add `changelogging` as a dependency with the following command:\n\n```console\n$ poetry add changelogging\n```\n\nOr by directly specifying it in the configuration like so:\n\n```toml\n[tool.poetry.dependencies]\nchangelogging = "^1.1.0"\n```\n\nAlternatively, you can add it directly from the source:\n\n```toml\n[tool.poetry.dependencies.changelogging]\ngit = "https://github.com/nekitdev/changelogging.git"\n```\n\n## Documentation\n\nYou can find the documentation [here][Documentation].\n\n## Support\n\nIf you need support with the library, you can send an [email][Email]\nor refer to the official [Discord server][Discord].\n\n## Changelog\n\nYou can find the changelog [here][Changelog].\n\n## Security Policy\n\nYou can find the Security Policy of `changelogging` [here][Security].\n\n## Contributing\n\nIf you are interested in contributing to `changelogging`, make sure to take a look at the\n[Contributing Guide][Contributing Guide], as well as the [Code of Conduct][Code of Conduct].\n\n## License\n\n`changelogging` is licensed under the MIT License terms. See [License][License] for details.\n\n[Email]: mailto:support@nekit.dev\n\n[Discord]: https://nekit.dev/discord\n\n[Actions]: https://github.com/nekitdev/changelogging/actions\n\n[Changelog]: https://github.com/nekitdev/changelogging/blob/main/CHANGELOG.md\n[Code of Conduct]: https://github.com/nekitdev/changelogging/blob/main/CODE_OF_CONDUCT.md\n[Contributing Guide]: https://github.com/nekitdev/changelogging/blob/main/CONTRIBUTING.md\n[Security]: https://github.com/nekitdev/changelogging/blob/main/SECURITY.md\n\n[License]: https://github.com/nekitdev/changelogging/blob/main/LICENSE\n\n[Package]: https://pypi.org/project/changelogging\n[Coverage]: https://codecov.io/gh/nekitdev/changelogging\n[Documentation]: https://nekitdev.github.io/changelogging\n\n[Discord Badge]: https://img.shields.io/badge/chat-discord-5865f2\n[License Badge]: https://img.shields.io/pypi/l/changelogging\n[Version Badge]: https://img.shields.io/pypi/v/changelogging\n[Downloads Badge]: https://img.shields.io/pypi/dm/changelogging\n\n[Documentation Badge]: https://github.com/nekitdev/changelogging/workflows/docs/badge.svg\n[Check Badge]: https://github.com/nekitdev/changelogging/workflows/check/badge.svg\n[Test Badge]: https://github.com/nekitdev/changelogging/workflows/test/badge.svg\n[Coverage Badge]: https://codecov.io/gh/nekitdev/changelogging/branch/main/graph/badge.svg\n',
    'author': 'nekitdev',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/nekitdev/changelogging',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



from pathlib import Path
from typing import Optional

import click
from iters import iter
from pendulum import Date, parse
from wraps import Option, wrap_optional

from changelogging import __version__ as version
from changelogging.build import Builder
from changelogging.config import Config
from changelogging.constants import DEFAULT_NAME, HASH, NEW_LINE
from changelogging.git import remove_paths
from changelogging.utils import today

__all__ = ("changelogging", "build", "create")


def get_config(string: Option[str]) -> Config:
    return Config.from_path(string.map_or_else(Path, Path))


def get_date(string: Option[str]) -> Date:
    return string.map_or_else(today, parse_date)


def parse_date(string: str) -> Date:
    return parse(string).date()  # type: ignore


split_lines = str.splitlines
right_strip = str.rstrip


@click.group(name=DEFAULT_NAME)
@click.help_option("-h", "--help")
@click.version_option(version, "-V", "--version")
def changelogging() -> None:
    pass


CONFIG_PATH = "config_path"
DATE_STRING = "date_string"


@changelogging.command()
@click.help_option("-h", "--help")
@click.option("-c", "--config", CONFIG_PATH, default=None)
@click.option("-d", "--date", DATE_STRING, default=None)
@click.option("-D", "--draft", is_flag=True, default=False)
@click.option("-r/-n", "--remove/--no-remove", default=False)
def build(
    config_path: Optional[str], date_string: Optional[str], draft: bool, remove: bool
) -> None:
    date = get_date(wrap_optional(date_string))

    config = get_config(wrap_optional(config_path))

    builder = Builder(config, date)

    if draft:
        click.echo(builder.build())

    else:
        if remove:
            remove_paths(builder.collect_paths())

        builder.write()


ABORTED = "Creation aborted."
CREATED = "Created the `{}` fragment."
PLACEHOLDER = "Add the content here."
EDIT = """
# Please enter the fragment content.
# Lines starting with `#` will be ignored.
# Close the file without saving to abort.
"""

NAME = "name"


def is_comment(line: str) -> bool:
    return line.startswith(HASH)


def is_content(line: str) -> bool:
    return not is_comment(line)


concat_new_line = NEW_LINE.join


@changelogging.command()
@click.help_option("-h", "--help")
@click.option("-c", "--config", CONFIG_PATH, default=None)
@click.option("-e/-n", "--edit/--no-edit", default=True)
@click.argument(NAME)
def create(config_path: Optional[str], edit: bool, name: str) -> None:
    config = get_config(wrap_optional(config_path))

    if edit:
        string = click.edit(EDIT)

        if string is None:
            click.echo(ABORTED)
            return

        string = iter(split_lines(string)).filter(is_content).map(right_strip).collect(concat_new_line)
    else:
        string = PLACEHOLDER

    path = config.directory / name

    path.write_text(right_strip(string) + NEW_LINE)

    click.echo(CREATED.format(name))
